import numpy as np
import collections

from scipy.special import binom
from scipy.interpolate import interp1d


def Bernstein(n, k):
    """Bernstein polynomial."""
    coeff = binom(n, k)

    def _bpoly(x):
        return binom(n, k) * x ** k * (1 - x) ** (n - k)

    return _bpoly


def Bezier(points, sample_points=200):
    """Build Bezier curve from points."""
    N = len(points)
    t = np.linspace(0, 1, sample_points)
    curve = np.zeros((sample_points, 2))
    for ii in range(N):
        curve += np.outer(Bernstein(N - 1, ii)(t), points[ii])
    return curve


"""
Adapted from https://github.com/reptillicus/Bezier
https://gist.github.com/Juanlu001/7284462
"""


class bezierMotionCurve:
    """
    Bezier curve with endpoints set at (0,0) and (1,1)
    """

    def __init__(self, x1, y1, x2, y2, sample_points=200):
        points = np.array([(0.0, 0), (x1, y1), (x2, y2), (1, 1)])
        x, y = Bezier(points, sample_points=sample_points).T

        self.func = interp1d(x, y, kind="quadratic")

    def __call__(self, t):
        return self.func(t)



from .colors import NamedColors
from .palettes import ColorLoversPalette, palette_blocks, palette_rectangles

from .interpolation import RGBa_interpolation
from .interpolation import LABa_interpolation

matplotlib_colors = NamedColors()
palette = ColorLoversPalette()



import numpy as np
import cv2
from ..artist import Artist, constant
import warnings


class gaussian_blur(Artist):
    blur_x = constant(0.2)
    blur_y = constant(0.2)

    args = ("blur_x", "blur_y")

    def draw(self, cvs, t=0.0):
        bx = cvs.transform_kernel_length(self.blur_x(t))
        by = cvs.transform_kernel_length(self.blur_y(t))
        kernel = (bx, by)
        cvs.img = cv2.GaussianBlur(cvs.img, kernel, 0)


class glow(Artist):

    glow_x = constant(2.0)
    glow_y = constant(1.0)
    n = constant(5)
    art = None

    args = ("art", "glow_x", "glow_y", "n")

    def draw(self, cvs, t=0.0):
        gx, gy = self.glow_x(t), self.glow_y(t)
        art = self.art(t)
        art.draw(cvs, t)

        if self.n(t) == 0:
            return True

        with cvs.layer() as cvs2:
            for i in range(self.n(t)):
                cvs2 += gaussian_blur(gx, gy)
                art.draw(cvs2, t)



import numpy as np
import cv2


def lerp(t, ys, xs=None):
    """
    Linear interpolation between the list of ys. If xs is not given
    points are evenly distributed between [0, 1]. Points are evaluated
    on the values of t.
    """
    if xs is None:
        xs = np.linspace(0, 1, len(ys))

    return np.interp(t, xs, ys)


def discrete_interpolation(t, alpha, cs):
    """
    No interpolation used, simply select the color on fixed intervals.
    Downweight by alpha.
    """
    cs = np.array(cs)
    n = len(cs)
    idx = np.clip(np.floor(t * n), 0, n - 1).astype(int)
    C = cs[idx, :].astype(np.float)
    C[:, 3] *= alpha
    return C


def RGBa_interpolation(t, alpha, cs):
    """
    Interpolate two colors RGB in RGB space parameterized by t. 
    Downweight by alpha.
    """
    csx = list(zip(*cs))
    R = lerp(t, csx[0])
    G = lerp(t, csx[1])
    B = lerp(t, csx[2])
    a = lerp(t, csx[3]) * alpha
    return np.vstack([R, G, B, a]).T


def LABa_interpolation(t, alpha, cs):
    """
    Interpolate two RGB colors in LAB space parameterized by t. 
    Downweight by alpha.
    """
    csx = list(zip(*cs))
    alpha_channel = csx[3]

    # Convert colors to LAB (remove alpha channel)
    C = np.array(cs).astype(np.uint8)[:, :3].reshape(-1, 1, 3)
    C = cv2.cvtColor(C, cv2.COLOR_RGB2LAB).reshape(-1, 3).T

    # Interpolate in LAB space
    L = lerp(t, C[0])
    A = lerp(t, C[1])
    B = lerp(t, C[2])

    # Convert back to RGB
    C = np.vstack([L, A, B]).T.reshape(-1, 1, 3)
    C = C.astype(np.uint8)
    C = cv2.cvtColor(C, cv2.COLOR_LAB2RGB).reshape(-1, 3)

    # Add back in the weighted alpha channel
    a = lerp(t, alpha_channel)
    a = np.clip(a * alpha, 0, 255).astype(np.uint8)

    # Stack them all together
    C = np.hstack([C, a.reshape(-1, 1)])

    return C



import itertools
import numpy as np
import cv2
from ..artist import Artist, constant, constant_list
from ..primitives import _DEFAULT_COLOR, _DEFAULT_SECONDARY_COLOR
from ..color import interpolation


class linear(Artist):
    colors = constant_list(_DEFAULT_COLOR, _DEFAULT_SECONDARY_COLOR)
    transparency = constant_list(None, None)
    theta = constant(0.0)
    interpolation = constant("LAB")

    args = ("colors", "transparency", "theta", "interpolation")

    def __call__(self, cvs, t=0.0, mask=None):
        """
        Assume the mask is of type Canvas. 
        Interpolation can be LAB or RGB.
        Theta controls the angle along the shape.

        If mask is None, return True to show that a gradient exists
        """
        if mask is None:
            return True

        mask_idx = mask.alpha > 0

        # If the mask_idx is 0, image is off the canvas
        if not mask_idx.sum():
            return True

        # Read/transform the colors, apply transparency if needed

        colors = []
        ITR = itertools.zip_longest(
            self.colors(t), self.transparency(t), fillvalue=None
        )

        for c, z in ITR:
            cval = cvs.transform_color(c).copy()
            if z is not None:
                cval[-1] = (np.clip(z, 0, 1) * 255).astype(np.uint8)
            colors.append(cval)

        theta = self.theta(t)
        A = np.array([np.cos(theta), np.sin(theta)])

        # Project each masked grid point onto the angle mapped by theta
        xg, yg = cvs.grid_points()
        B = np.vstack([xg[mask_idx], yg[mask_idx]])

        pro = A.dot(B)
        pro -= pro.min()
        pro /= pro.max()

        # Smooth the image based off the alpha from the mask image
        alpha = (mask.alpha / 255.0)[mask_idx]

        imode = self.interpolation(t)
        if imode == "LAB":
            C = interpolation.LABa_interpolation(pro, alpha, colors)
        elif imode == "RGB":
            C = interpolation.RGBa_interpolation(pro, alpha, colors)
        elif imode == "discrete":
            C = interpolation.discrete_interpolation(pro, alpha, colors)
        else:
            raise KeyError(f"Unknown interpolation {imode}")

        # Blend the new shape in
        rhs = cvs.copy()
        rhs.img[mask_idx] = C
        cvs.blend(rhs)



"""
Base easing functions modified from: https://easings.net/
"""

import math
import numpy as np
import copy
from .bezier import bezierMotionCurve
from scipy.interpolate import interp1d


class EasingBase:
    def __init__(self, start=0, stop=1, translate=0.0, scale=1.0):
        self.start = start
        self.stop = stop
        self.translate = 0.0
        self.scale = 1.0

    def copy(self):
        # Returns a deep copy of the Easing
        return copy.deepcopy(self)

    def __neg__(self):
        rhs = self.copy()
        rhs.scale *= -1
        return rhs

    def __add__(self, val):
        rhs = self.copy()
        rhs.translate += val
        return rhs

    def __radd__(self, val):
        return self + val

    def __sub__(self, val):
        rhs = self.copy()
        rhs.translate -= val
        return rhs

    def __mul__(self, val):
        rhs = self.copy()
        rhs.scale *= val
        return rhs

    def __rmul__(self, val):
        return self * val

    def __truediv__(self, val):
        rhs = self.copy()
        rhs.scale /= val
        return rhs

    def __call__(self, t):
        a = self.func(t)

        value = self.start * (1 - a) + self.stop * a
        return self.scale * (self.translate + value)

    def func(self, t):
        raise NotImplementedError


class BezierEase(EasingBase):
    def __init__(self, start=0, stop=1, flip=None, phase=0.0, *args, **kwargs):
        super().__init__(start, stop, *args, **kwargs)

        # Lazy loading of the Bezier curve so we can quickly create objects
        # note that x0, y0, x1, y1 must be set in the derived class
        self.f = None

        self.phase = phase
        self.flip = flip

        if flip is True:
            self.flip = interp1d([0, 0.5, 1.0], [0, 1, 0])

    def get_params(self):
        """
        Return a set of arguments to construct the easing.
        """
        return (self.x0, self.y0, self.x1, self.y1)

    def func(self, t):
        t += self.phase

        # If t is STRICTLY greater than 1, take the mod
        if isinstance(t, np.ndarray):
            t[t > 1] = np.mod(t[t > 1], 1.0)
            t[t < 0] = np.mod(t[t < 0], 1.0)
        else:
            if t > 1 or t < 0:
                t %= 1.0

        if self.f is None:
            self.f = bezierMotionCurve(self.x0, self.y0, self.x1, self.y1)

        if self.flip:
            t = self.flip(t)

        return self.f(t)


#########################################################################
# Named Easing functions
#########################################################################


class Linear(EasingBase):
    def func(self, t):
        return t


class easeInSine(BezierEase):
    x0, y0, x1, y1 = 0.47, 0, 0.745, 0.715


class easeOutSine(BezierEase):
    x0, y0, x1, y1 = 0.39, 0.575, 0.565, 1


class easeInOutSine(BezierEase):
    x0, y0, x1, y1 = 0.445, 0.05, 0.55, 0.95


class easeInQuad(BezierEase):
    x0, y0, x1, y1 = 0.55, 0.085, 0.68, 0.53


class easeOutQuad(BezierEase):
    x0, y0, x1, y1 = 0.25, 0.46, 0.45, 0.94


class easeInOutQuad(BezierEase):
    x0, y0, x1, y1 = 0.455, 0.03, 0.515, 0.955


class easeInCubic(BezierEase):
    x0, y0, x1, y1 = 0.55, 0.055, 0.675, 0.19


class easeOutCubic(BezierEase):
    x0, y0, x1, y1 = 0.215, 0.61, 0.355, 1


class easeInOutCubic(BezierEase):
    x0, y0, x1, y1 = 0.645, 0.045, 0.355, 1


class easeInQuart(BezierEase):
    x0, y0, x1, y1 = 0.895, 0.03, 0.685, 0.22


class easeOutQuart(BezierEase):
    x0, y0, x1, y1 = 0.165, 0.84, 0.44, 1


class easeInOutQuart(BezierEase):
    x0, y0, x1, y1 = 0.77, 0, 0.175, 1


class easeInQuint(BezierEase):
    x0, y0, x1, y1 = 0.755, 0.05, 0.855, 0.06


class easeOutQuint(BezierEase):
    x0, y0, x1, y1 = 0.23, 1, 0.32, 1


class easeInOutQuint(BezierEase):
    x0, y0, x1, y1 = 0.86, 0, 0.07, 1


class easeInExpo(BezierEase):
    x0, y0, x1, y1 = 0.95, 0.05, 0.795, 0.035


class easeOutExpo(BezierEase):
    x0, y0, x1, y1 = 0.19, 1, 0.22, 1


class easeInOutExpo(BezierEase):
    x0, y0, x1, y1 = 1, 0, 0, 1


class easeInCirc(BezierEase):
    x0, y0, x1, y1 = 0.6, 0.04, 0.98, 0.335


class easeOutCirc(BezierEase):
    x0, y0, x1, y1 = 0.075, 0.82, 0.165, 1


class easeInOutCirc(BezierEase):
    x0, y0, x1, y1 = 0.785, 0.135, 0.15, 0.86


class easeInBack(BezierEase):
    x0, y0, x1, y1 = 0.6, -0.28, 0.735, 0.045


class easeOutBack(BezierEase):
    x0, y0, x1, y1 = 0.175, 0.885, 0.32, 1.275


class easeInOutBack(BezierEase):
    x0, y0, x1, y1 = 0.68, -0.55, 0.265, 1.55


#########################################################################
# Custom Easing functions
#########################################################################



from contextlib import contextmanager
import cv2
import os
import numpy as np
import collections
from . import Artist
from .color import matplotlib_colors


class Canvas:
    """
    Basic canvas object for pixelhouse drawings. 
    Extent measures along the x-axis.
    """

    def __init__(
        self,
        width=200,
        height=200,
        extent=4.0,
        bg="black",
        img=None,
        name="pixelhouseImage",
        shift=8,
    ):

        self.shift = shift

        # Assign the background color, but make sure it is fully transparent
        # needed for antialiased edges
        bg = np.array(self.transform_color(bg)).astype(np.uint8)
        self.bg = bg

        if img is None:
            # Empty canvas
            self._img = np.zeros((height, width, 4), np.uint8)
            self.bg[3] = 0
            self.img = self.bg

        else:
            # If a single channel image, upcast to greyscale
            if len(img.shape) == 2:
                img = np.dstack([img, img, img])

            # If rgb, add the alpha channel
            if img.shape[2] == 3:
                alpha = np.zeros_like(img[:, :, 0])
                img = np.dstack([img, alpha])
            self._img = img

        self.name = name
        self.extent = extent

        # When animating, we may want to pass attributes from one canvas
        # to another. Do so in the shared_attributes
        self.shared_attributes = {}

    def __repr__(self):
        return (
            f"Canvas (w/h) {self.height}x{self.width}, " f"extent {self.extent}"
        )

    @property
    def height(self):
        return self.img.shape[0]

    @property
    def width(self):
        return self.img.shape[1]

    @property
    def channels(self):
        return self.img.shape[2]

    @property
    def aspect_ratio(self):
        return self.width / float(self.height)

    @property
    def xmin(self):
        return self.inverse_transform_x(0)

    @property
    def xmax(self):
        return self.inverse_transform_x(self.width)

    @property
    def ymin(self):
        return self.inverse_transform_y(0)

    @property
    def ymax(self):
        return self.inverse_transform_y(self.height)

    @property
    def shape(self):
        return self.height, self.width, self.channels

    @property
    def pixels_per_unit(self):
        return self.width / (2 * self.extent)

    @property
    def img(self):
        """
        Returns the image
        """
        return self._img

    @img.setter
    def img(self, value):
        """
        Set the img from either a scalar, vector, or image array.
        """
        if isinstance(value, str):
            value = self.transform_color(value)

        value = np.array(value)

        rank = np.ndim(value)
        if not rank:
            self._img.fill(value)
            return
        elif rank == 1:
            self._img[:, :] = self.transform_color(value)
        else:
            self._img = value

    @property
    def rgb(self):
        """
        Return the image without the alpha channel
        """
        return self.img[:, :, :3]

    @rgb.setter
    def rgb(self, value):
        """
        Set the alpha channel for the image, 
        either as a fixed value or an array.
        """
        self.img[:, :, :3] = value

    @property
    def alpha(self):
        """
        Return the alpha channel from the image
        """
        return self.img[:, :, 3]

    @alpha.setter
    def alpha(self, value):
        """
        Set the alpha channel for the image, 
        either as a fixed value or an array.
        """
        self.img[:, :, 3] = value

    def blank(self, bg=None):
        """
            Return an empty canvas of the same size.
            Does not modify in place.
        """
        if bg is None:
            bg = self.bg

        cvs = Canvas(self.width, self.height, extent=self.extent, bg=bg)
        return cvs

    def copy(self, bg=None, transparent=False):
        """
            Returns a deep copy of this canvas
        """
        cvs = self.blank()
        cvs.img = self.img.copy()

        if transparent:
            cvs.alpha = 0

        return cvs

    def __call__(self, art=None):
        """
            Calls an Artist on the canvas.
            Example: canvas(translate(x=2, y=1))
        """
        if art is not None:
            art.draw(self)
        return self

    def __getitem__(self, key):
        return self.img[key]

    def __setitem__(self, key, val):
        self.img[key] = val

    def __len__(self):
        """
        Return 2 so calling functions work seamlessly between 
        Canvas and Animation (allows for interpolation to not complain).
        """
        return 2

    def combine(self, rhs, mode="blend"):
        if rhs.width != self.width:
            raise ValueError("Can't combine images with different widths")

        if rhs.height != self.height:
            raise ValueError("Can't combine images with different heights")

        if mode == "blend":
            self.blend(rhs)
        elif mode == "add":
            cv2.add(self.img, rhs.img, self.img)
        elif mode == "subtract":
            cv2.subtract(self.img, rhs.img, self.img)
        else:
            raise ValueError(f"Unknown mode {mode}")

        return self

    def __iadd__(self, rhs):
        """
        Add an Artist to the canvas, or combine two canvas depending
        on what the rhs is.
        """
        if isinstance(rhs, Artist):
            self(rhs)
        elif isinstance(rhs, Canvas):
            self.combine(rhs)
        else:
            raise TypeError(f"Can't combine a Canvas with a {type(rhs)}")

        return self

    def cv2_draw(self, func, args, mode, **kwargs):

        if mode == "direct":
            func(self.img, *args)
        elif mode == "gradient":
            rhs = self.blank()
            func(rhs.img, *args)
            kwargs["gradient"](self, t=kwargs["t"], mask=rhs)
        else:
            rhs = self.blank()
            func(rhs.img, *args)
            self.combine(rhs, mode)

        return self

    def blend(self, rhs):
        # Smooth the image based off the alpha from the mask image

        a = rhs.alpha.reshape(self.height, self.width, -1).astype(np.float32)
        a /= 255

        MX = (1 - a) * self.img + a * rhs.img
        MX = np.clip(MX, 0, 255).astype(np.uint8)
        self.img = MX

        return self

    def transform_x(self, x, is_discrete=True, use_shift=False):
        x += self.extent
        x *= self.pixels_per_unit

        if is_discrete:
            if use_shift and self.shift:
                x *= 2 ** self.shift
            return int(x)

        return x

    def inverse_transform_x(self, x):
        x /= self.pixels_per_unit
        x -= self.extent
        return x

    def transform_length(self, r, is_discrete=True, use_shift=False):
        r *= float(self.width) / self.extent / 2

        if use_shift and self.shift:
            r *= 2 ** self.shift

        if is_discrete:
            return int(r)
        return r

    def transform_y(self, y, is_discrete=True, use_shift=False):
        y *= -1
        y += self.extent / self.aspect_ratio
        y *= self.pixels_per_unit

        if is_discrete:
            if use_shift and self.shift:
                y *= 2 ** self.shift
            return int(y)

        return y

    def inverse_transform_y(self, y):
        y /= self.pixels_per_unit
        y -= self.extent / self.aspect_ratio
        y *= -1
        return y

    def transform_kernel_length(self, r):
        # Kernels must be positive and odd integers
        r = self.transform_length(r, is_discrete=False)

        remainder = r % 2
        r = int(r - r % 2)
        r += -1 if remainder < 1 else 1

        r = max(1, r)
        return r

    def transform_thickness(self, r):
        # If thickness is negative, leave it alone
        if r > 0:
            return self.transform_length(r)

        return r

    def transform_color(self, c):
        """
        Cleanly transform a color.
        """

        if isinstance(c, str):
            c = matplotlib_colors(c)

        # Force add in the alpha channel
        if len(c) == 3:
            c = list(c) + [255]

        # Force the return of a scalar (cv2 is picky)
        if isinstance(c, np.ndarray):
            c = c.tolist()

        return c

    def grid_coordinates(self):

        attrs = self.shared_attributes

        if (
            "grid_coordinates" in attrs.keys()
            and attrs["prior_shape"] == self.shape
        ):
            return attrs["grid_coordinates"]

        shape = tuple(self.shape)

        xg, yg, zg = np.meshgrid(
            np.arange(shape[1]), np.arange(shape[0]), np.arange(shape[2])
        )

        attrs["grid_coordinates"] = (xg, yg, zg)
        attrs["prior_shape"] = self.shape

        return attrs["grid_coordinates"]

    def grid_points(self):

        attrs = self.shared_attributes

        if "grid_points" in attrs.keys() and attrs["prior_shape"] == self.shape:
            return attrs["grid_points"]

        xg, yg, _ = self.grid_coordinates()
        xg = xg[:, :, 0]
        yg = yg[:, :, 0]

        xp = self.inverse_transform_x(xg.astype(np.float32))
        yp = self.inverse_transform_y(yg.astype(np.float32))

        attrs["grid_points"] = (xp, yp)
        attrs["prior_shape"] = self.shape

        return attrs["grid_points"]

    @staticmethod
    def transform_angle(rads):
        # From radians into degrees, counterclockwise
        return -rads * (360 / (2 * np.pi))

    @staticmethod
    def get_lineType(antialiased):
        if antialiased:
            return cv2.LINE_AA
        return 8

    def show(
        self, delay=0, orient=False, return_status=False
    ):  # pragma: no cover

        """ 
        Opens a preview window displaying the image. 
        Returns the keycode of the button pressed if return_status==True,
        else returns self.
        """

        # Before we show we have to convert back to BGR
        dst = cv2.cvtColor(self.img, cv2.COLOR_RGB2BGR)

        cv2.imshow(self.name, dst)

        if orient:
            cv2.moveWindow(self.name, 40, 40)

        status = cv2.waitKey(delay)

        if return_status:
            return status

        return self

    def save(self, filename):

        """ Writes the canvas out to the specified filename
        """

        # Before we save we have to convert back to BGR
        dst = cv2.cvtColor(self.img, cv2.COLOR_RGB2BGR)
        cv2.imwrite(filename, dst)

        return self

    def load(self, filename):

        """ Reads an image in from a specified filename
            and converts it to RGB space
        """

        if not os.path.exists(filename):
            raise FileNotFoundError(f'"{filename}" not found')

        # Read the image in and convert to RGB space
        img = cv2.imread(filename, cv2.IMREAD_UNCHANGED)

        # If needed, add in an alpha channel as fully opaque
        if img.shape[2] == 3:
            alpha = 255 * np.ones_like(img[:, :, 0])
            img = np.dstack((img, alpha))

        self.img = cv2.cvtColor(img, cv2.COLOR_BGRA2RGBA)

        return self

    @contextmanager
    def layer(self):
        canvas = self.copy(transparent=True)
        yield canvas
        self += canvas

    def resize(self, fx=0, fy=0, output_size=None):
        """
        Resizes the canvas either by a set scale or direct pixel amount
        """

        if output_size and (fx or fy):
            raise ValueError("Can't set both fx,fy and output_size")

        if output_size:
            self.img = cv2.resize(self.img, output_size)
        else:
            if not fy:
                fy = fx
            self.img = cv2.resize(self.img, (0, 0), fx=float(fx), fy=float(fy))

        return self


def hstack(canvas_list):
    """
    Stacks the images from two or more canvas horizontally,
    uses the extent and background color from the first canvas on the list.
    """
    cvs = canvas_list[0].blank()
    cvs.img = np.hstack([x.img for x in canvas_list])
    return cvs


def vstack(canvas_list):
    """
    Stacks the images from two or more canvas vertically,
    uses the extent and background color from the first canvas on the list.
    """
    cvs = canvas_list[0].blank()
    cvs.img = np.vstack([x.img for x in canvas_list])
    return cvs


def gridstack(canvas_grid):
    """
    Stacks the an array of arrays of canvas into a grid.
    Uses the extent and background color from the first canvas on the list.
    """
    cvs = canvas_grid[0][0].blank()
    rows = [hstack([x for x in row]) for row in canvas_grid]
    return vstack(rows)


def load(filename):
    """
    Helper function to load an image directly. 
    pixelhouse.load(filename) returns a preloaded canvas.
    """
    return Canvas().load(filename)
